/*



*/

/**
   This program analyzes an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

       Prog ::= Exp
              | '(' 'prog' (Fun | Exp)* Exp ')'

        Fun ::= '(' 'fun' VARIABLE Lambda ')'   // a function declaration

     Lambda ::= '(' 'lambda' VARIABLE* Exp ')'  // formal parameters followed by function body

        Exp ::= Apply
              | If
              | While
              | Begin
              | Var
              | Set
              | Print
              | BExp
              | AExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Apply ::= '(' 'apply' Exp Exp* ')'  // function value followed by actual parameters

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp ')'

      Begin ::= '(' 'begin' Exp* Exp ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

        Set ::= '(' 'set' VARIABLE Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [-][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class StaticAnalysis
{
   public static int DEBUG = 1;

   public static Environment globalEnv; // reference to the global environment object
                                        // this is needed to implement lexical scope
   /**



   */
   public static boolean analyze(Tree tree)
   {
      // Instantiate a global environment object.
      globalEnv = new Environment();

      return analyzeProg( tree, globalEnv );
   }//analyze()


   // Analyze a prog.
   private static boolean analyzeProg(Tree tree, Environment env)
   {
      boolean result = true;

      // Check which kind of Prog we have.
      if ( ! tree.getElement().equals("prog") )
      {
         // Analyze the single expression.
         result = analyzeExp(tree, env);
      }
      else
      {
         boolean rv = true;
         // Analyze each Fun and gather all the function
         // names in the global environment. Do this before
         // analyzing each function's lambda expression so
         // that we can allow function forward references
         // (and mutually recursive functions).
         for (int i = 0; i < tree.degree(); ++i)
         {
            if ( tree.getSubTree(i).getElement().equals("fun") )
            {
               rv = analyzeFun(tree.getSubTree(i), env);
            }
            result = result && rv;
         }
         // Analyze each Fun or Exp in the Prog. In this pass,
         // we analyze the body (lambda expression) of each
         // function definition.
         for (int i = 0; i < tree.degree(); ++i)
         {
            if ( tree.getSubTree(i).getElement().equals("fun") )
            {
               if ( env.defined( tree.getSubTree(i).getSubTree(0).getElement() )
                 && env.lookUp( tree.getSubTree(i).getSubTree(0).getElement() ).tag.equals("lambda") )
               {
                  rv = analyzeLambda(tree.getSubTree(i).getSubTree(1), env);
               }
            }
            else
            {
               rv = analyzeExp(tree.getSubTree(i), env);
            }
            result = result && rv;
         }
      }

      return result;
   }//analyzeProg()


   // Analyze the first part of a function definition
   // (without analyzing the lambda part).
   // This is needed by the REPL, so we make it public.
   public static boolean analyzeFun(Tree tree, Environment env)
   {
      boolean result = true;



      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//analyzeFun()


   // Analyze a lambda expression.
   // This is needed by the REPL, so we make it public.
   public static boolean analyzeLambda(Tree tree, Environment env)
   {
      boolean result = true;




//      if (DEBUG > 0) System.out.println( tempEnv + "\n" ); // for debugging purposes

      return result;
   }//analyzeLambda()


   // Analyze an expression.
   public static boolean analyzeExp(Tree tree, Environment env)
   {
      boolean result = true;

      String node = tree.getElement();

      if ( node.equals("apply") )
      {
         result = analyzeApply(tree, env);
      }
      else if ( node.equals("if") )
      {
         result = analyzeIf(tree, env);
      }
      else if ( node.equals("while") )
      {
         result = analyzeWhile(tree, env);
      }
      else if ( node.equals("begin") )
      {
         result = analyzeBegin(tree, env);
      }
      else if ( node.equals("var") )
      {
         result = analyzeVar(tree, env);
      }
      else if ( node.equals("set") )
      {
         result = analyzeSet(tree, env);
      }
      else if ( node.equals("print") )
      {
         result = analyzePrint(tree, env);
      }
      else if ( node.equals("&&")
             || node.equals("||")
             || node.equals("!") )
      {
         result = analyzeBExp(tree, env);  // boolean expression
      }
      else if ( node.equals("<")  || node.equals(">")
             || node.equals("<=") || node.equals(">=")
             || node.equals("==") || node.equals("!=") )
      {
         result = analyzeRExp(tree, env);  // relational operator
      }
      else if ( node.equals("+") || node.equals("-")
             || node.equals("*") || node.equals("/")
             || node.equals("%") || node.equals("^") )
      {
         result = analyzeAExp(tree, env);  // arithmetic expression
      }
      else if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = true;
         }
         else if ( node.matches("^[-]*[0-9][0-9]*") )
         {
            result = true;
         }
         else if ( env.defined(node) )  // a variable
         {
            result = true;
         }
         else
         {
            result = false;
            System.err.println("**** Analysis Error: undefined variable: "
                               + node
                               + " at line " + tree.getToken().line
                               + ", position " + tree.getToken().position
                               + "\n" );
         }
      }
      else
      {
         result = false;
         System.err.println("**** Analysis Error: illegal expression: "
                            + node + ":"
                            + " at line " + tree.getToken().line
                            + ", position " + tree.getToken().position
                            + "\n" );
      }

      return result;
   }//analyzeExp()


   // Analyze a function application.
   private static boolean analyzeApply(Tree tree, Environment env)
   {
      boolean result = true;




      return result;
   }//analyzeApply()


   // Analyze an if expression.
   private static boolean analyzeIf(Tree tree, Environment env)
   {
      boolean result = true;



      return result;
   }//analyzeIf()


   // Analyze a while expression.
   private static boolean analyzeWhile(Tree tree, Environment env)
   {
      boolean result = true;



      return result;
   }//analyzeWhile()


   // Analyze a begin (block) expression.
   private static boolean analyzeBegin(Tree tree, Environment env)
   {
      boolean result = true;



      return result;
   }//analyzeBegin()


   // Analyze a variable declaration expression.
   private static boolean analyzeVar(Tree tree, Environment env)
   {
      boolean result = true;




      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//analyzeVar()


   // Analyze a variable assignment expression.
   private static boolean analyzeSet(Tree tree, Environment env)
   {
      boolean result = true;



      return result;
   }//analyzeSet()


   // Analyze a print expression.
   private static boolean analyzePrint(Tree tree, Environment env)
   {
      boolean result = true;



      return result;
   }//analyzePrint()


   // Analyze a boolean expression.
   private static boolean analyzeBExp(Tree tree, Environment env)
   {
      boolean result = true;



      return result;
   }//analyzeBExp()


   // Analyze a relational expression.
   private static boolean analyzeRExp(Tree tree, Environment env)
   {
      boolean result = true;



      return result;
   }//analyzeRExp()


   // Analyze an arithmetic expression.
   private static boolean analyzeAExp(Tree tree, Environment env)
   {
      boolean result = true;


      return result;
   }//analyzeAExp()


   /**
      Returns true if the given string is one of the reserved keywords.
   */
   private static boolean isKeyword(String word) {
      return keywords.contains(word);
   }

   private static final java.util.List<String> keywords
                   = java.util.Arrays.asList(
                       "prog",
                       "fun",
                       "lambda",
                       "apply",
                       "if",
                       "while",
                       "begin",
                       "var",
                       "set",
                       "print",
                       "true",
                       "false"
                     );
}//StaticAnalysis
